
;****************************************************************************
; Ju + Te Tiny  4k EMR-ES 1988  Teil 2
; (c) Dr. Helmut Hoyer, 1987-1989
;
; reass: Patrick Schfer 10/2011
;****************************************************************************

	cpu z8601
	page 0
	include	stddefz8.inc

ROMsize		SET 2048
TapeInMask	SET 1		; original ist P3.0 Kassetteneingang
;TapeInMask	SET 2		; P3.1 damit P3.0 frei ist fr SIO

; AS-Funktionen
hi	function x,(x>>8)&255
lo	function x, x&255

; Adressen im unteren ROM

TBS		EQU 00824h
ZBS4		EQU 00872h
CLS		EQU 008DDh
Begin		EQU 008FFh
RB_PrintFile	EQU 00A62h
NLE		EQU 00ACEh
NLN		EQU 00AD4h
IRQ4_Handler	EQU 00ADAh
PrtChar		EQU 00B95h
PrtTxt1		EQU 00BFAh


;****************************************************************************
;  06.05.2006 (js)
;
;  Speicherbelegung, 2k-System
;
;      +-----------+-----------------------------------+
;      |1k RAM FFFF| 512 Byte Graphik-                 |
;      |       FE00|          Bildwiederholspeicher    |
;      +-----------+-----------------------------------+
;      |       FDFF| 128 Byte Systemstack              |
;      |       FD80|                                   |
;      +-----------+-----------------------------------+
;      |       FD7F| 128 Byte ASCII-                   |
;      |       FD00|          Bildwiederholspeicher    |
;      +-----------+-----------------------------------+
;      |       FCFF| 128 Byte frei fr Standard-       |
;      |       FC80|          Programme                |
;      +-----------+-----------------------------------+
;      |       FC7F| 128 Byte Magnetbandpuffer         |
;      |       FC00|                                   |
;      +-----------+-----------------------------------+
;
;      +-----------+-----------------------------------+
;      |       FBFF| Anwenderbereich                   |
;      |       E000|                                   |
;      +-----------+-----------------------------------+
;
;      +-----------+-----------------------------------+
;      |2k ROM 27FF| 2048 Byte Betriebssystem          |
;      |Modul2 2000|           4K-System Teil 2        |
;      +-----------+-----------------------------------+
;
;      +-----------+-----------------------------------+
;      |2k ROM 0FFF| 2048 Byte Betriebssystem          |
;      |Modul1 0800|                                   |
;      +-----------+-----------------------------------+
;      |2k ROM 07FF| 2048 Byte Basicinterpreter        |
;      |intern 0000|                                   |
;      +-----------+-----------------------------------+
;
;
;  19.10.2011 (ps)
;
;  Registerbelegung
;
; %02-03 Arbeitsregister X
; %04-05 Arbeitsregister Y
; %06-07	BASIC-Startadresse, default %E000
; %08-09 Adresse der Prozedurtabelle. %08=00 wenn interne verwendet
; %0E		BASIC-Fehler
; 		Fehlerbits: 	0..3 GOSUB Rekursionstiefe
;                 		4  GOSUB Stackberlauf (01)
;                 		5  RETURN ohne GOSUB (02)
;                 		6 Division durch Null (04)
;                 		7 Integerberlauf (08)
; %0F		BASIC-Flags
;		02: Programm beendet
;		04:
;		08: STOP aktiv
; fr BASIC-Editor
; %10-11	Pointer auf BASIC-Programmzeile
; %12-13	temp	z.B. Wort fr Dezimalausgabe
; 			%13 Zeichen von getchar
; %15-19	temp	z.B. ASCII-Stellen fr Dezimalausgabe
; 			%15 Zeichen an putchar
; %20-21	BASIC-Variable A
; %22-23	BASIC-Variable B
;   (...)
; %4E-4F	BASIC-Variable X
; %50-51	BASIC-Variable Y oder Programmstart fr SAVE
; %52-53	BASIC-Variable Z oder Programmende fr SAVE
; fr Betriebsystem
; %52-53	Anfang der Markentabelle
; %54		Videozeilenzhler
; %58		temp	z.B. Textnummer
; %5A		ASCII-bergaberegister
; %5B		Cursorposition im Zeichenspeicher (0..127)
; %5C		temp	z.B. Schrittweite beim Cursor
; %5D		HexByte-bergaberegister
; %5E-5F	Adress-bergaberegister
; %60		temp
; %6C		Byte aus Hex-Eingabe ber Tastatur
; %6D		Tastencode-Register. T-Bit (bit 7) =1 wenn neuer Inhalt
; %6E-6F	Wort aus Hex-Eingabe ber Tastatur
; %74-75	IRQ0-Vektor
; %76-77	IRQ1-Vektor
; %78-79	IRQ2-Vektor
; %7A-7B	IRQ3-Vektor
; %7C-7D	IRQ4-Vektor (T0 fr Videoerzeugung)
; %7E-7F	IRQ5-Vektor
;
;
;  Speicherbelegung
;
; %FFFF		Shift-Flag. 0 wenn Shift aktiv
;               Das gibt einen weissen Strich unten rechts auf dem Bildschirm.
;
;
;  BASIC
:
; 10 PRINT "HALLO"	E000: 80 0A 50 22 48 41 4C 4C 4F 22 0D
; 20 GOTO 10		E00B: 80 14 47 31 30 0D
;			E011: 00      
;
; d.h. eine Zeile beginnt mit 2 byte Zeilennr binr, oberstes bit gesetzt
; dann Befehlstoken bzw ASCII, dann %0D als Zeilenende.
; Semikolon trennt Anweisungen, Komma trennt Argumente
;
;****************************************************************************

	org 2000h

ES_Begin	di     
		srp #0F0h		; SFR-Page
	assume RP:0F0h
		ld SPL, #000h		; Stackpointer auf %FE00
		ld SPH, #0FEh
		ld IMR, #10h		; IRQ4 erlauben, INTs aus
		ld IPR, #8		; IRQ4 auf niedrigste Prioritt
		ld P01M, #92h		; P0/P1 Adressbus, Stack extern, Timing normal
		ld PRE0, #7		; Prescaler =1, Timer modulo N
		ld T0, #64		; Timer 0 auf 64 (d.h. INT4 alle 64s bzw 15,625kHz)
		ld TMR, #3		; Timer 0 laden und starten
		clr 6Dh			; Tastaturpuffer lschen
		srp #060h
	assume RP:060h
		ld R12, #4
		ld 7Ch, #hi(IRQ4_Handler)	; Vektor fr T0-Interrupt
		ld 7Dh, #lo(IRQ4_Handler)
ES_BgLp		ld 58h, R12		; Text 4..1 fr Startmen
		call PrtText3		;  am Bildschirm ausgeben
		djnz R12, ES_BgLp

		ei     
		call TBS		; auf Tastendruck warten
		cp 5Ah, #'B'
		jp Z, Begin		; B --> BASIC-Start
		cp 5Ah, #'S'
		jr NZ, ES_BgNotS
		call ES_Save1		; S --> Save, Magnetbandausgabe
ES_BgNotS	cp 5Ah, #'L'
		jr NZ, ES_BgNotL
		call ES_Load1		; L --> Load, Magnetbandeingabe
ES_BgNotL	cp 5Ah, #'R'
		jr NZ, ES_BgNotR
		call ES_Data		; R --> RAM- und Registeranzeige
ES_BgNotR	cp 5Ah, #'P'
		jp Z, ES_Prog		; P --> Programm-Modus
		cp 5Ah, #'I'
		jr NZ, ES_Begin		; Neustart wenn nicht I

; I--> Init, also Speicher lschen
		call TBS
		cp 5Ah, #'!'		; shift-1 ?
		jr NZ, ES_Begin		;  nein --> Abbruch, Neustart

		di     			; die Glotze aus, dann gehts schneller
		ld R0, #0FDh
		ld R1, #0		; Start bei %FBFF
		ld R2, #0		; mit 0 berschreiben
		ld R3, #8		; Ende bei %0900
ES_InitLp	decw RR0
		lde @RR0, R2
		cp R3, R0
		jr C, ES_InitLp
		jr ES_Begin

;-----------------------------------------------------------------------------
; Texttabelle fr Textausgabe PrtText3
;-----------------------------------------------------------------------------

TextTab2	db 'P'+80h, "ROG"				; #01
		db 0Dh+80h, "  P"				; #02
		db 'D'+80h, "ATA"				; #03
		db 0Ch+80h, " EMR-ES 1988", 0Dh, " ", 0Dh 
		db "  B BASIC",0Dh, "  I INIT", 0Dh
		db "  S SAVE", 0Dh, "  L LOAD", 0Dh, "  R"	; #04
		db 0Ch+80h, "MTB"				; #05


;-----------------------------------------------------------------------------
; Hexadezimalanzeige eines Bytes aus %5D, Kursor (%5B) wird aktualisiert
;-----------------------------------------------------------------------------

ES_PrHex	call ES_PrHex1		; Funktion 2x ausfhren

ES_PrHex1	swap 5Dh
		ld 5Ah, 5Dh
ES_PrHex3	and 5Ah, #0Fh		; oberes Nibble maskieren
		cp 5Ah, #10		; Buchstabe?
		jr C, ES_PrHex2
		add 5Ah, #7		; dann noch 7 dazuaddieren
ES_PrHex2	add 5Ah, #030h
		jp ZBS4			; Zeichen ausgeben


;-----------------------------------------------------------------------------
; Hexadezimalanzeige der Speicherzelle mit der Adresse aus %5E und %5F, 
; Adresse wird um 1 erhht, der Kursor (%5B) wird aktualisiert
;-----------------------------------------------------------------------------

ES_PrMemByt	push RP
		srp #050h
	assume RP:050h
		ld R12, #05Dh		; 5Ch, also eins vor der auszugebenden Adr
		ldei @R12, @RR14	; Wert holen, R12 ist 5Ch
		pop RP
		jr ES_PrHex		; und ausgeben


;-----------------------------------------------------------------------------
; Hexadezimalanzeige eines Doppelbytes aus %5E und %5F, 
; Kursor (%5B) wird aktualisiert
;-----------------------------------------------------------------------------

	assume RP:0F0h			; falscher Wert um lange Zuweisungen zu erzwingen
ES_PrMemWrd	ld 5Dh, 5Eh
		call ES_PrHex		; erst das erste Byte..
		ld 5Dh, 5Fh
		jr ES_PrHex		; ... dann das zweite


;-----------------------------------------------------------------------------
; Berechnen der Byteanzahl in %6B eines Befehls mit dem Operationscode aus %5D
;-----------------------------------------------------------------------------

ES_GetOpLen	ld 6Bh, #3
		cp 5Dh, #0D4h		; CALL iRR?
		jr Z, ES_GetOpLen2	;  --> hat nur 2 bytes !
		ld 6Ah, 5Dh
		and 6Ah, #0Fh		; untere 4 bit maskieren
		cp 6Ah, #14		; grergleich 14?
		jr NC, ES_GetOpLen1	; --> xE, xF hat 1 byte
		cp 6Ah, #12		; gleich 12?
		jr Z, ES_GetOpLen2	; --> xC hat 2 byte
		tm 6Ah, #4		; bit 4 gesetzt?
		jr NZ, ES_GetOpLen3	; --> x4-7, xD hat 3 byte
ES_GetOpLen2	dec 6Bh			; Rest hat 2 byte Lnge
		ret  
ES_GetOpLen1	ld 6Bh, #1
ES_GetOpLen3	ret  
  

;-----------------------------------------------------------------------------
; Textausgabe. Die Nummer des Textes steht in Register 58h
; Die Texttabelle ist TextTab2
;-----------------------------------------------------------------------------

PrtText3	push RP
		srp #050h
	assume RP:050h
		ld R14, #hi(TextTab2)	; Zeiger auf Texttabelle setzen
		ld R15, #lo(TextTab2)
		jp PrtTxt1		; und Text ausgeben, Return von da



;****************************************************************************
; Markenverwaltung 
;
; Es gibt Sprungziel-Marken "ABC:" links vom Befehl, und Absprung-Marken
; ":XYZ" als Operand. Eingegeben werden sie jedoch immer mit dem Doppelpunkt
; am Anfang, also im Operanden-Format.
;
; Die Tabelle wchst von oben nach unten. Jeder Eintrag ist 6 Bytes lang.
; %52/53 zeigt auf den Anfang ("MTB").
;
;   %E000    -->       MTB
;   %64   65   66 67 68 69
;   AdrH AdrL  A  B  C  :    --> Sprungziel ABC:
;   AdrH AdrL  :  X  Y  Z    --> Absprung (Operand) :XYZ
;****************************************************************************

; Speicheradresse zu einer Marke bestimmen.
; R0/R1 zeigt auf den Eintrag in der Markentabelle
; R2/R3 enthlt die aufgelste Adresse

	assume RP:060h
MGetLabel	sub R1, #4		; Markenname berspringen
		sbc R0, #0
		lde R3, @RR0		; LSB holen
		decw RR0
		lde R2, @RR0		; MSB holen
		add R1, #5		; Zeiger wieder zurckstellen
		adc R0, #0
		ret    


; Abstand der Speicheradresse in RR4 zur aktuellen Marke bestimmen,
; Ergebnis in RR2. Z=1 heisst wir sind auf der Marke, C=1 heisst die Marke
; liegt im Speicher hinter uns.

MFndAdr3	call MGetLabel		; Speicheradresse zur aktuellen Marke
		sub R3, R5		; minus aktuelle Adresse
		sbc R2, R4
		ret    


; Speicheradresse aus R4/R5 in der Markentabelle suchen
; C=1, Z=1 : als Sprungziel gefunden, RR0 zeigt auf den Eintrag
; C=1, Z=0 : als Absprung (Operand) gefunden, RR0 zeigt auf den Eintrag
; C=0, Z=1 : Ende der Tabelle erreicht

MFindAdr	ld R0, 52h		; Anfang der Markentabelle
		ld R1, 53h
MFndAdr2	lde R2, @RR0		; erstes Byte holen
		or R2, R2
		jr Z, MFndAdr1		; 0 --> Ende

		call MFndAdr3		; Speicheradresse auflsen und prfen
		or R2, R3		; sind wir das?
		jr Z, MFndAdr1		; gefunden --> Ende

MFndAdr5	sub R1, #6
		sbc R0, #0
		jr MFndAdr2		; nchste Marke prfen

MFndAdr1	rcf    
		lde R2, @RR0		; erstes Zeichen der Marke
		or R2, R2		; Null?
		jr Z, MFndAdr4		;  --> Tabellenende, Z=1
		cp R2, #':'
		scf    
MFndAdr4	ret    


; prfen, ob die Marke in (%66)..(%69) in der Markentabelle enthalten ist.
; (%52/53) enthlt Zeiger auf die Markentabelle
; Carry gelscht wenn gefunden, RR0 zeigt auf den Eintrag
; Carry gesetzt wenn nicht gefunden

MFindLabel	ld R0, 52h		; Beginn der Markentabelle
		ld R1, 53h
MFndLbl4	ld R2, #69h		; eingegebene Marke
MFndLbl2	lde R3, @RR0
		cp R3, @R2
		jr NZ, MFndLbl1		; erstes Zeichen ungleich, weiter
		decw RR0		; sonst nchstes prfen
		dec R2
		cp R2, #66h
		jr NC, MFndLbl2

		add R1, #4		; Zeiger auf Fundstelle
		adc R0, #0		; Carry gelscht
		ret    

MFndLbl1	or R3, R3		; Ende erreicht?
		jr NZ, MFndLbl3		;  nein -->
MEntLbl1	scf    			; Carry setzen
		ret 
   
MFndLbl3	sub R2, #63h		; #66h-#63h = 3
		sub R1, R2		; Zeiger auf nchsten Markeneintrag
		sbc R0, #0
		jr MFndLbl4		; und weitersuchen


; Marke (in %66..69, %66 enthlt ':') wurde als Operand eingegeben.
; Prfen, ob die Adresse in RR4 (%64/65) in der Markentabelle enthalten ist.  
; Falls nicht, dann als Absprung-Marke eintragen

MEnterOpr	call MFindAdr		; gibt's die Adresse in der Markentabelle?
MEntOpr1	jr NC, MNewEntry	;  nein --> eintragen
		jr NZ, MNewEntry	;  ja, als Sprungziel --> eintragen
		call MFndAdr5		;  ja, als Absprung --> weitersuchen
		jr MEntOpr1

; neue Marke (in %66..69) in Tabelle eintragen
MNewEntry	sub R1, #5
		sbc R0, #0		; Zeiger auf neues Element in der Tabelle
		ld R2, #6		
		ld R3, #64h		; %64/65 Adresse, %66..69 Name
MNewEnt1	ldei @RR0, @R3		
		djnz R2, MNewEnt1
MLinkPgm1	ret    


; Marke (in %66..69, %69 enthlt ':') wurde als Sprungziel angegeben
; Prfen ob Marke bereits in der Markentabelle enthalten ist.
; Falls ja, Carry setzen und zurck. Ziele darf es nur einmal geben.
; Falls nicht, dann als Sprungziel-Marke eintragen
; (um ein Sprungziel mit einem neuen Wert zu belegen, muss der zugehrige
; Befehl mit 'X' gelscht werden)

MEnterLabel	call MFindLabel		; gibt's den Namen in der Markentabelle
		jr NC, MEntLbl1		; jawoll, Carry setzen und zurck
		call MFindAdr		; gibt's die Adresse in RR4 schon in der Tabelle
MEntLbl2	jr Z, MNewEntry		; ja, als Absprung --> eintragen
		call MFndAdr5		; sonst weitersuchen
		jr MEntLbl2


; Binden (Linken) mit Offset in %70/71
; Fr jede Absprung-Marke in der Markentabelle wird das gleichnamige 
; Sprungziel gesucht. Abbruch mit C=1 wenn keins gefunden. 
; Die Absprung-Marke zeigt auf den Befehl. Hinter dem wird bei
; relativen Sprngen der Abstand zwischen Sprungziel und Absprung
; eingetragen, und bei absoluten Sprngen das Sprungziel plus Offset.
; C=0 wenn alles erfolgreich durchlaufen wurde.

MLinkPgm	ld R10, 52h		; Anfang der Markentabelle
		ld R11, 53h

MLinkPgm3	ld R0, R10
		ld R1, R11
		call MFndAdr1		; erste Marke prfen
		jr NC, MLinkPgm1	; Abbruch bei Tabellenende (C=0, Z=1)
		jr Z, MLinkPgm2		; Sprungziel (C=1, Z=1) --> nchster Eintrag

		ld R2, #3
		ld R3, #66h
		ld R9, #':'		; %69 = ':'
		decw RR0
		decw RR0
MLinkPgm4	ldei @R3, @RR0		; Markenname nach %66..68 bertragen
		djnz R2, MLinkPgm4

		call MFindLabel		; und ein Sprungziel dieses Namens finden
		jr C, MLinkPgm1		;  existiert nicht --> Abbruch
		call MGetLabel		; sonst Adresse auflsen, Ergebnis in RR2
		ld R0, R10		; Adresse der Absprung-Marke zurckholen
		ld R1, R11
		ld R4, R2		; RR4 = Sprungziel absolut
		ld R5, R3
		call MGetLabel		; und ebenfalls auflsen
		lde R6, @RR2		; Befehl an der Absprungadresse
		tm R6, #4		; bit 3 gelscht ?
		jr Z, MLinkPgm5		;  --> relativer Sprung

		add R5, 71h		; Offset zu Zieladresse hinzuaddieren
		adc R4, 70h
		incw RR2
		lde @RR2, R4		; MSB nach dem Befehl eintragen,
		jr MLinkPgm6		;  dann LSB

MLinkPgm5	sub R5, R3		; LSB Sprungziel - Absprungadresse
		sub R5, #2		;  -2 gibt relativen Abstand
MLinkPgm6	incw RR2		; nach dem Befehl eintragen
		lde @RR2, R5

MLinkPgm2	sub R11, #6		; nchste Marke
		sbc R10, #0
		jr MLinkPgm3

;****************************************************************************


; drei aufeinanderfolgende Nullbytes suchen (Programm-Ende)
; RR6 zeigt auf das dritte Nullbyte

FindPgmEnd	ld R6, R4		; Ziel = Start
		ld R7, R5
FindPgmEnd1	ld R10, #-4
FindPgmEnd2	incw RR6		; Ziel +1
		lde R11, @RR6		; Byte von Zieladresse holen	
		or R11, R11		; Null ?
		jr NZ, FindPgmEnd1	;  nein, nchste Adresse
		inc R10			;  ja, R10 erhhen 
		jr NZ, FindPgmEnd2	;   und nchste Adresse
		ret    


; Ein Byte an Adresse RR4 einfgen, Adressen in der Markentabelle korrigieren

InsertByte	call FindPgmEnd		; letzte benutzte Speicherstelle suchen
		ld R10, R6	
		ld R11, R7
		decw RR10
		ld R8, R6		; Endadresse
		ld R9, R7
		sub R9, R5		; minus Startadresse
		sbc R8, R4		;  gibt Lnge des zu verschiebenden Blocks
InsertByt1	lde R12, @RR10		; vorletztes Byte holen
		lde @RR6, R12		; und auf letzte Stelle schreiben
		decw RR6	
		decw RR10
		decw RR8	
		jr NZ, InsertByt1

		ld R10, #0FFh		; -1 Byte Offset fr Einfgen
		ld R11, #0FFh		;  (beim Lschen +1..+3)
		lde @RR4, R10		; neue Stelle mit %FF belegen

DelByt3		ld R0, 52h		; Anfang der Markentabelle
		ld R1, 53h
InsertByt4	call MFndAdr3		; Abstand akt. Adresse zur Marke 
		jr C, InsertByt2	;  negativ, Marke liegt hinter uns

		sub R1, #4
		sbc R0, #0		; Zeiger auf Adresse
		lde R3, @RR0		; LSB holen
		sub R3, R11		;  Offset abziehen
		lde @RR0, R3		;  und zurckschreiben
		decw RR0
		lde R2, @RR0		; MSB holen
		sbc R2, R10		;  bertrag verarbeiten
		lde @RR0, R2		;  und zurckschreiben
		decw RR0		; Zeiger auf nchsten Eintrag

InsertByt3	call MFndAdr1		; Tabellenende erreicht? 
		jr C, InsertByt4	;  nein, nchste Marke prfen
		ret    

InsertByt2	sub R1, #6		; nchste Marke
		sbc R0, #0
		jr InsertByt3


; Maschinenbefehl, d.h. 1..3 Byte, an Adresse RR4 lschen,
; Eintrge fr RR4 aus der Markentabelle lschen, Adressen korrigieren.

DelBytes	call FindPgmEnd
		lde R12, @RR4		; zu lschendes Befehlsbyte
		ld 5Dh, R12
		call ES_GetOpLen	; Lnge ermitteln
		push R11		;  und sichern
		add R11, R5		; RR10 = Startadresse + Lnge
		clr R10			;  gibt erstes zu verschiebendes Byte
		adc R10, R4
		sub R7, R11		; Programmende minus erstes zu
		sbc R6, R10		;  verschiebendes Byte gibt Zhler
		ld R8, R4		; Zieladresse
		ld R9, R5
DelByt1		lde R12, @RR10		; Byte von Quelladresse
		lde @RR8, R12		; zur Zieladresse
		incw RR10
		incw RR8
		decw RR6
		jr NZ, DelByt1

DelByt6		call MFindAdr		; gelschte Adresse in der Markentabelle?
		jr NC, DelByt2		; nein, restliche Adressen korrigieren

		ld R8, R0		; Adresse des zu lschenden Eintrags
		ld R9, R1
		ld R10, R0
		ld R11, R1
		sub R11, #6		; Adresse des folgenden Eintrags
		sbc R10, #0
DelByt5		ld R6, #6
DelByt4		lde R12, @RR10		; folgende Marke
		lde @RR8, R12		;  auf aktuelle Position kopieren
		decw  RR8
		decw RR10
		djnz R6, DelByt4

		lde R12, @RR8
		or R12, R12		; Tabellenende erreicht?
		jr NZ, DelByt5		; nein, weiter mit nchstem Eintrag
		jr DelByt6

DelByt2		ld R10, #0
		pop R11			; Befehlslnge als Offset
		jr DelByt3		;  und Markentabelle anpassen


; Zeichen von Tastatur holen und in %5A zurckliefern
; Carry gesetzt wenn auerhalb '0'..'F' oder ':'
; Zero gesetzt wenn ':'

ES_GtHxDigit	call TBS		; auf Tastendruck warten
		cp 5Ah, #'0'		; <030h ?
		jr C, ES_GtHxDigit1	;   --> Fehler!
		cp 5Ah, #'F'		; <=046h ?
		jr ule, ES_GtHxDigit2	;   --> ok
ES_GtHxDigit1	scf    			; Carry setzen, zurck
		ret    
ES_GtHxDigit2	cp 5Ah, #':'		; =':' ?
		jr Z, ES_GtHxDigit1	;   --> Zeroflag gesetzt
		rcf    			; Carry lschen, zurck
		ret    


; Wert des Eingabezeigers (%70/71) am Bildschirm ausgeben, die folgenden 6
; Felder mit Leerzeichen berschreiben, dann zur Hexeingabe bergehen

ES_GetHxData	srp #050h
	assume RP:050h
		ld R11, #070h
		ld R14, 70h		; Eingabezeigerwert (%70/71) 
		ld R15, 71h		;  nach %5E/5F kopieren
		call ES_PrMemWrd	;  und drucken
		inc R11			; = %71
		srp #060h
	assume RP:060h
		ld R12, #6		
		ld 5Ah, #' '
ES_GetHxDta1	call ZBS4		; 6 Leerzeichen drucken
		djnz R12, ES_GetHxDta1
		sub 5Bh, #6		; und Cursor wieder zurckstellen


;-----------------------------------------------------------------------------
; Hexadezimaleingabe eines Bytes in %6C mit Tastatur und Anzeige RP = #%60
; Carry gelscht wenn Wert gltig, gesetzt wenn Eingabefehler
; Zero gesetzt wenn ':'
;-----------------------------------------------------------------------------

ES_GetHexByt	call ES_GtHxDigit	; ein Zeichen holen
		jr C, ES_GetHxByt2	;  keine Hexziffer wenn C=1
		ld R12, 5Ah		; sonst in %6C zwischenparken
		call ES_GtHxDigit	; und zweites Zeichen holen
		jr NC, ES_GetHxByt1	;  auch ok --> weiter
		or R12, R12		; Zeroflag lschen (':' nur als erstes Zeichen erlaubt)
ES_GetHxByt2	ret  

ES_GetHxByt1	call ES_GetHxByt3	; zweiten ASCII in Hexwert wandeln
		ld R3, 5Ah
		ld 5Ah, R12
		call ES_GetHxByt3	; ersten ASCII wandeln
		ld R12, 5Ah
		swap R12		; nach oben
		or R12, R3		; und kombinieren
		rcf    			; Carry lschen
		ret    

; ASCII in %5A in Hexwert 0..Fh wandeln
ES_GetHxByt3	cp 5Ah, #':'
		jr C, ES_GetHxByt4	; kleiner, also Ziffer
		add 5Ah, #9		; Buchstaben 041h..046h --> 04Ah..04Fh
ES_GetHxByt4	and 5Ah, #0Fh		; unteres Nibble ausmaskieren
		ret    


; restliche Kommandos der RAM- und Registeranzeige, werden vom Programm-
; Modus mitverwendet

ES_Dta13	cp 5Ah, #'+'		; "+"?
		jr NZ, ES_Dta14
		incw rr14		; Eingabezeiger erhhen
ES_GetHxWrd2	ret    

ES_Dta14	cp 5Ah, #'G'		; "G"?
		jr NZ, ES_GetHxWrd2	; nein, zurcl
; sonst neuen Wert fr Anfangsadresse abholen


;-----------------------------------------------------------------------------
; Hexadezimaleingabe eines Doppelbytes in %6E und %6F mit Tastatur und Anzeige RP = #%60
; Carry gelscht wenn Wert gltig, gesetzt wenn Eingabefehler
;
; Falls erstes Zeichen ':' wird eine 3stellige Textmarke erwartet und der
; zugehrige Adresswert aus der Markentabelle zurckgeliefert.
;-----------------------------------------------------------------------------

ES_GetHexWrd	call ES_GetHexByt	; ein Byte holen
		jr NC, ES_GetHxWrd1	;  ok, Hexwert --> weiter
		jr NZ, ES_GetHxWrd2	; kein ':' --> abbrechen

; Textmarke eingeben
		ld R12, #3		; drei...
ES_GetHxWrd3	call TBS		;  ... Zeichen holen...	
		push 5Ah		;  ... und auf Stack zwischenlagern
		djnz R12, ES_GetHxWrd3

		ld R12, #3
		ld R9, #068h
ES_GetHxWrd4	pop @r9			; vom Stack nach %66..%68
		dec R9
		djnz R12, ES_GetHxWrd4

		ld R9, #':'		; %69=':'
		call MFindLabel		; Marke in Markentabelle vorhanden?
		jr C, ES_GetHxWrd2	;  --> nein, Abbruch
		call MGetLabel		; sonst Adresse auflsen
		ld R14, R2		; %6E
		ld R15, R3		; %6F
		ret   

; zweites Byte holen und zurck 
ES_GetHxWrd1	ld R14, R12		; erstes Byte %6C nach %6E
		call ES_GetHexByt	; nchstes Byte holen
		jr C, ES_GetHxWrd2	; Fehler --> abbrechen
		ld R15, R12		;  sonst zweites Byte %6C nach %6F
		ret    


;-----------------------------------------------------------------------------
; RAM- und Registeranzeige
; Anfangsadresse in %6E/6F
;-----------------------------------------------------------------------------

ES_Data		call CLS
		push RP
ES_Dta3		srp #060h
	assume RP:060h
		ld 70h, R14		; Eingabezeiger auf Anfangsadresse setzen
		ld 71h, R15		;  (%70/71)            (%6E/6F)
ES_Dta9		srp #050h
	assume RP:050h
		ld R8, #3		; Text #3 "DATA"
		call PrtText3		; am Bildschirm ausgeben
		ld R14, 6Eh		; Startadresse auf Anfangsadresse
		ld R15, 6Fh		;   (%5E/5F)
		call ES_PrMemWrd	; Adresse drucken
		call NLE
		srp #060h
	assume RP:060h

		ld R12, #6		; 6 Zeilen
ES_Dta2		ld R11, #4		;  mit je 4 Bytes
		ld 5Ah, 5Fh		; LSB der aktuellen Adresse
		call ES_PrHex3		; unterstes Nibble drucken
ES_Dta1		ld 5Ah, #' '
		call ZBS4		; Leerzeichen ausgeben
		or R14, R14		; ist das MSB = 0 ?
		jr Z, ES_Dta11		;  --> dann sind wir im Registerbereich

		call ES_PrMemByt	; RAM-Inhalt %5E/5F drucken, Adresse inkr.
		jr ES_Dta12

ES_Dta11	ld 5Dh, @5Fh		; Registerinhalt %5F drucken,
		inc 5Fh			;  Adresse inkrementieren
		call ES_PrHex

ES_Dta12	djnz R11, ES_Dta1	; nchstes Byte
		djnz R12, ES_Dta2	; nchste Zeile

		call ES_GetHxData	; Eingabezeigeradresse drucken, Byte abholen
		jr C, ES_DtaK		; C=1 --> keine Hexzahl, Kommandos prfen

		srp #070h
	assume RP:070h
		or R0, R0		; MSB des Eingabezeigers = 0 ?
		jr Z, ES_Dta10		;  --> Register !

		ld R2, #6Ch		; Hexwert in die Speicheradresse
		ldei @RR0, @r2		;  am Eingabezeiger schreiben
		jr ES_Dta8

ES_Dta10	ld @r1, 6Ch		; Hexwert ins Register schreiben
		incw RR0

ES_Dta8		clr 5Bh			; Cursor auf Anfang
		jr ES_Dta9

	assume RP:060h
ES_DtaK		cp 5Ah, #060h		; "Off"?
		jr C, ES_Dta4		
		pop  RP			; ja, RAM- und Registeranzeige beenden
		ret 
   
ES_Dta4		cp 5Ah, #0Dh		; "Enter"?
		jr NZ, ES_Dta5
		ld R13, #0
		add R15, #4		; Startadresse um 4 erhhen
		adc R14, R13
		jr ES_Dta6		; und neu starten

ES_Dta5		cp 5Ah, #'-'		; "-"?
		jr NZ, ES_Dta7
		decw 70h		; Eingabezeiger dekrementieren
		ld R4, 70h
		ld R5, 71h
		sub R5, R15
		sbc R4, R14		; vor Startadresse?
		jr NC, ES_Dta8		;  nein, --> weiter
		ld R14, 70h		;  sonst Anfangsadresse auf Eingabezeiger
		ld R15, 71h
		jr ES_Dta8

ES_Dta7		call ES_Dta13		; Kommandos "+" und "G" prfen

ES_Dta6		clr 5Bh			; Cursor auf Anfang
		jp ES_Dta3		;  und neu starten


;-----------------------------------------------------------------------------
; Programm-Modus
;-----------------------------------------------------------------------------

ES_Prog		push FLAGS
		push RP
		srp #050h
	assume RP:050h
		ld R8, #5		; Text #5, "MTB"
		call PrtText3	
		ld R14, R2		; Adresse der Markentabelle (%5E/5F)
		ld R15, R3
		call ES_PrMemWrd	;  anzeigen
		ld R11, #4
		srp #060h
	assume RP:060
		call ES_GetHexWrd	; neuen Wert holen
		jr C, ES_Prog1		; Fehler --> berspringen
		ld 52h, R14		; sonst Wert setzen
		ld 53h, R15
ES_Prog1	pop RP

ES_Prog4	di     			; Video aus

; "PROG" und den Inhalt des Flagregisters anzeigen
		call CLS
		ld 58h, #1		; Text #1, "PROG"
		call PrtText3
		pop 6Ch			; Flags 
		push 6Ch
		push RP
		srp #050h
	assume RP:050h
		ld R15, #8		; 8 Flag-Bits
ES_Prog2	ld R10, #'0'
		rl 6Ch			
		adc R10, R8		; ASCII '0' oder '1'
		call ZBS4		;  ausgeben
		djnz R15, ES_Prog2

		ld R15, 6Fh		; Anfangsadresse auf Eingabewert
		ld R14, 6Eh		;  fr Markentabelle setzen
		clr 70h
		lde R13, @rr14		; erstes Befehlsbyte holen (R13 = %5D)
		call ES_GetOpLen	;  und Lnge des Befehls ermitteln
		ld 71h, R15		; aktuelle Adresse...
		add 71h, 6Bh		;  ... plus Befehlslnge in %6B...
		adc 70h, R14		;  ... gibt Adresse des zweiten Befehls in %70/71

; sechs Zeilen Befehlsbytes anzeigen
ES_Prog3	call ES_PrtPgmLn	; Befehlszeile ab RR14 (%5E/5F) ausgeben
	assume RP:060h			; (wird in ES_PrtPgmLn umgestellt)
		cp  5Bh, #070h		; Cursor in der letzten Zeile?
		jr C, ES_Prog3		;  nein --> nchste Befehlszeile bearbeiten

		ei     			; Video ein

; *** Eingabezeile entgegennehmen
ES_Prog6	call ES_GetHxData	; Adresse in %70/71 ausgeben und Eingabe holen
		jr NC, ES_ProgH1	;  gltige Hexzahl --> Wert bernehmen
		jp NZ, ES_ProgK1	;  kein ':' --> auf Kommando prfen

; * Eingabe war ':', also Sprungziel-Marke erwarten
		ld  5Bh, #073h		; Cursor auf Eingabeposition zurck
		call PrtChar		;  und Zeichen drucken
		ld  5Bh, #070h		; Cursor zum Zeilenanfang
		ld R12, #3		; die Marke hat 3 Byte
ES_ProgM1	call TBS		; auf Tastendruck warten
		push 5Ah		;  und Eingabe auf Stack
		djnz R12, ES_ProgM1

		ld R12, #3
		ld R9, #68h
ES_ProgM2	pop @R9			; dann nach %66..68
		dec R9
		djnz R12, ES_ProgM2

		ld  R9, #':'		; %69 = ':'
		ld  R4, 70h		; Adresse des ersten Befehlsbytes
		ld  R5, 71h
		call MEnterLabel	; Sprungziel-Marke eintragen
		jr NC, ES_Prog5		;  ok, fertig
		jr ES_Prog6		;  Namen gibt's schon, neu versuchen!

; * Eingabe war Hexzahl, also neuen Befehl beginnen
ES_ProgH1	srp #070h
	assume RP:070h
		ld R3, 6Ch		; Hex-Eingabewert nach %73 kopieren,
		ld 5Dh, R3		; als neues Befehlsbyte nehmen und
		lde @RR0, R3		;  in (%70/71) speichern
		srp #060h
	assume RP:060h
		call ES_GetOpLen	; Lnge des Befehls ermitteln
		ld R14, 70h
		ld R15, 71h
		djnz R11, ES_ProgH2	; falls Lnge >1 ...

ES_Prog5	pop RP
		jp ES_Prog4		; fertig, nchste Runde

ES_ProgH2	call ES_GetHexByt	; nchstes Byte des Befehls holen
		jr NC, ES_ProgH3	;  gltige Hexzahl --> weiter
		jr NZ, ES_Prog5		;  kein ':' --> Abbruch, nchste Runde

; knnen wir berhaupt Marken als Operand gebrauchen?
		cp 73h, #0D6h		; CALL ?
		jr Z, ES_ProgM5		;  --> ok
		and 73h, #0Fh
		cp 73h, #0Ch		; LD ?
		jr Z, ES_Prog5		;  --> nicht ok, Abbruch
		cp 73h, #0Ah		; JR, JP, INC oder 0xFh?
		jr C, ES_Prog5		;  --> nein, Abbruch

; also eine Marke eingeben
ES_ProgM5	ld R12, #3
ES_ProgM3	call TBS
		push 5Ah		; erstmal auf den Stack...
		djnz R12, ES_ProgM3

		ld R12, #3
		ld R6, #69h
ES_ProgM4	pop @r6			; ...dann nach %67..69
		dec R6
		djnz R12, ES_ProgM4

		ld R4, 70h		; Adresse des ersten Befehlsbytes
		ld R5, 71h
		ld R6, #':'
		call MEnterOpr		; Absprungmarke eintragen
		jr ES_Prog5		; fertig

; Hexzahl direkt bernehmen
ES_ProgH3	incw 70h		; Adresszhler inkrementieren
		srp #070h
	assume RP:070h
		ld  R3, 6Ch
		lde @RR0, R3		; und Byte einsortieren
		dec 6Bh			; Lnge -1
		jr Z, ES_Prog5		; fertig --> Ende
		srp #060h
	assume RP:060h
		call ES_GetHexByt	; sonst nchstes Byte holen
		jr ES_ProgH3		;  und weitermachen

; * Eingabe auf Kommando prfen
ES_ProgK1	cp 5Ah, #'Q'		; Q: Eingabe in Zeile 1 ?
		jr NZ, ES_ProgK2	;  nein, weiter
		ld 70h, R14		; aktuelle Adresse auf Anfangsadresse
		ld 71h, R15
ES_Prog7	jp ES_Prog6		; und neue Eingabe entgegennehmen

ES_ProgK2	cp 5Ah, #'-'		; -: Adresse -1 ?
		jr NZ, ES_ProgK3	;  nein, weiter
		decw 6Eh		; Anfangsadresse -1
		clr 6Dh			; aktuellen Befehl ungltig machen
ES_ProgK7	jr ES_Prog5		

ES_ProgK3	cp 5Ah, #'T'		; T: Binden auf Adresse ?
		jr NZ, ES_ProgK4	;  nein, weiter
		push R14 		; Anfangsadresse sichern
		push R15
		call ES_GetHexWrd	; Zieladresse abfragen
		clr 70h
		clr 71h
		pop R5			; Anfangsadresse in RR4
		pop R4
		jr C, ES_ProgK16	; Eingabefehler --> Adresse beibehalten!
		sub R15, R5		; Offset berechnen, d.h.
		sbc R14, R4		; neue Adr - Anfangsadr
		ld 70h, R14
		ld 71h, R15
ES_ProgK16	ld R14, R4		; Anfangsadresse wiederherstellen
		ld R15, R5
		di     			; Video aus
		call MLinkPgm		; Marken auflsen mit Offset %70/71
		jp NC, ES_Prog5		; ok, fertig

		ei     			; sonst Video ein zur Fehlerausgabe
		ld R9, #'?'
		push R9
		ld R12, #3
		ld R9, #68h
ES_ProgK5	push @R9		; '?'+Markenname auf Stack
		dec R9
		djnz R12, ES_ProgK5

		ld 5Bh, #075h		; Cursor in die unterste Zeile
		ld R12, #4
ES_ProgK6	pop 5Ah
		call ZBS4		; und die Zeichen vom Stack ausgeben
		djnz R12, ES_ProgK6
		call NLN
		jr ES_Prog7		; und neue Eingabe entgegennehmen

ES_ProgK4	cp 5Ah, #'I'		; I: Byte einfgen ?
		ld R4, 70h		; aktuelle Adresse in %64/65
		ld R5, 71h
		jr NZ, ES_ProgK8	;  nein, weiter
		call InsertByte
		jr ES_ProgK7

ES_ProgK8	cp 5Ah, #'X'		; X: Befehl streichen ?
		jr NZ, ES_ProgK9
		call DelBytes
		jr ES_ProgK7

ES_ProgK9	lde R11, @RR14		; Befehl auf Anfangsadresse holen
		ld 5Dh, R11
		call ES_GetOpLen	; und Lnge bestimmen
		cp 5Ah, #0Dh		; Enter: nchste Zeile anzeigen ?
		jr NZ, ES_ProgK10	;  nein, weiter
		clr R13
		add R15, R11		; sonst Anfangsadresse
		adc R14, R13		;   um Befehlslnge erhhen
		jr ES_ProgK7

ES_ProgK10	cp 5Ah, #'N'		; N: Schleifen-Test ?
		jr NZ, ES_ProgK11	;  nein, weiter
		clr 70h
		ld 71h, R11		; Befehlslnge
		add 71h, R15		; plus Anfangsadresse
		adc 70h, R14		;  gibt Haltepunkt
		jr RunBrkPoint1		

ES_ProgK11	cp 5Ah, #'H'		; H: Haltepunkt auf Adresse ?
		jr NZ, ES_ProgK12	;  nein, weiter
		push 6Eh		; Anfangsadresse sichern
		push 6Fh
		call ES_GetHexWrd	; Haltepunkt abfragen
		ld 70h, R14
		ld 71h, R15
		pop 6Fh			; Anfangsadresse wiederherstellen
		pop 6Eh
		jp  C, ES_ProgK7	; Abbruch bei Fehleingabe
RunBrkPoint1	pop  RP
		jr RunBrkPoint		; Programm ausfhren bis Adr %70/71

ES_ProgK12	pop  RP
		cp 5Ah, #'L'		; L: Echtzeit-Test ?
		jr NZ, ES_ProgK13	;  nein, weiter
		pop FLAGS
		jp @6Eh			; Sprung zur Anfangsadresse

ES_ProgK13	cp 5Ah, #'R'		; R: Aufruf DATA ?
		jr NZ, ES_ProgK14	;  nein, weiter
		push 6Eh		; Anfangsadresse sichern
		push 6Fh
		clr 6Eh			; und auf RP setzen
		ld >6Fh, RP
		call ES_Data		; RAM - und Registeranzeige aufrufen
		pop 6Fh			; Anfangsadresse wiedeherstellen
		pop 6Eh
ES_ProgK17  	jp ES_Prog4

ES_ProgK14	cp 5Ah, #060h		; OFF: Rcksprung
		jr C, ES_ProgK15	;  nein, weiter
		jp ES_Begin

ES_ProgK15  	cp 5Ah, #'S'		; S: Schritt-Test ?
		jr Z, RunBrkPoint2	; 
		push RP
		srp #060h
	assume RP:060h
		call ES_Dta13		; + und G testen
		jp ES_Prog5


; Programm in Einzelschritten ausfhren bis zur Adresse in %70/71
; Falls %70 = 0 wird nur ein Schritt ausgefhrt.
; Programmstart ist die Anfangsadresse (%6E/6F), diese wird aktualisiert.

RunBrkPoint2	clr 70h			; Haltepunkt %00xx heisst Single-Step

RunBrkPoint	di     
		and TMR, #0FCh		; Timer T0 aus
		ld 7Ch, #hi(RunBrkPtInt)	; IRQ4-Vektor
		ld 7Dh, #lo(RunBrkPtInt)
		and IRQ, #02Fh		; anstehenden T0-Interrupt lschen
		ld T0, #4		; T0 auf 4 s
RunBrkPoint4	pop FLAGS
		push 6Fh		; Anfangsadresse und FLAGS auf Stack
		push 6Eh		;  d.h. einen Interrupt simulieren
		push FLAGS
		or TMR, #3		; Timer starten
		iret   			; und simulierten Interrupt beenden
; jetzt luft das Programm ab der Anfangsadresse, bis der Timer zuschlgt
; Der IRQ4-Einsprung ist genau hier.
RunBrkPtInt	and TMR, #0FCh		; Timer T0 aus
		pop FLAGS
		pop 6Eh			; Anfangsadresse aktualisieren
		pop 6Fh
		push FLAGS
		and IRQ, #47
		or 70h, 70h		; single-step ?
		jr Z, RunBrkPoint3	;  ja, also fertig
		cp 70h, 6Eh		; Endadresse erreicht?
		jr NZ, RunBrkPoint4	;  nein, nchster Befehl
		cp 71h, 6Fh
		jr NZ, RunBrkPoint4
RunBrkPoint3	ld 7Ch, #hi(IRQ4_Handler)	; IRQ4-Vektor zurckstellen
		ld 7Dh, #lo(IRQ4_Handler)
		ld T0, #64		; T0 wieder auf 64s
		or TMR, #3		;  und einschalten
		jr ES_ProgK17


;-----------------------------------------------------------------------------
; Anzeige einer Maschinenprogrammzeile mit der Adresse aus %5E und %5F, 
; Adresse und Kursor (%5B) werden aktualisiert, RP := #%60
; Wenn die Adresse in der Markentabelle enthalten ist, wird stattdessen die
; Marke ausgegeben.
;-----------------------------------------------------------------------------

ES_PrtPgmLn	srp #060h
	assume RP:060h
		ld R4, 5Eh		; Adresse des Befehls
		ld R5, 5Fh

; falls Adresse in der Markentabelle, Marke drucken ("ABC:")
		call MFindAdr		; in der Markentabelle vorhanden?
ES_PPL3		jr NC, ES_PPL1		;  nein --> weiter

		jr Z, ES_PPL2		; als Sprungziel ?
		call MFndAdr5		;  nein --> nchste prfen
		jr ES_PPL3

ES_PPL2		ld R12, #4		; vier Zeichen
ES_PPL4		lde R2, @RR0		; aus Markentabelle holen
		push R2			;  auf Stack zwischenlagern
		decw RR0
		djnz R12, ES_PPL4
		
		ld R12, #4		; vier Zeichen
ES_PPL5		pop 5Ah			; von Stack holen
		call ZBS4		;  und ausgeben
		djnz R12, ES_PPL5
		jr ES_PPL6

; ohne Marke nur Hex-Adresse drucken
ES_PPL1		call ES_PrMemWrd	

; erstes Befehlsbyte ausgeben
ES_PPL6		srp #050h
	assume RP:050h
		ld R10, #' '
		call ZBS4		; Leerzeichen drucken
		lde R13, @RR14		; Befehl holen
		call ES_PrHex		;  und ausgeben, Adresse erhhen
		call ES_GetOpLen	; Lnge bestimmen
		ld R10, #' '
		call ZBS4		; Leerzeichen drucken

; falls Operandenadresse in der Markentabelle, Marke drucken (":XYZ")
		srp #060h
	assume RP:060h
		ld R4, 5Eh		
		ld R5, 5Fh
		call MFindAdr		; nchste Adresse in der Markentabelle?
ES_PPL9		jr NC, ES_PPL7		; nein --> weiter

		jr NZ, ES_PPL8		; als Operand ?
		call MFndAdr5		;  ja --> nchste Marke prfen
		jr ES_PPL9

ES_PPL8		add 5Fh, 6Bh		; Befehlslnge zur Adresse addieren
		adc 5Eh, #0

		ld R12, #4		; vier Zeichen
ES_PPL10	lde R2, @RR0		; aus Markentabelle holen
		push R2			;  auf Stack zwischenlagern
		decw RR0
		djnz R12, ES_PPL10

		ld R12, #4		; vier Zeichen
ES_PPL11	pop 5Ah			; von Stack holen
		call ZBS4		;  und ausgeben
		djnz R12, ES_PPL11
		jr ES_PPL12

; sonst nur die Bytes ausgeben
ES_PPL7		incw 5Eh		; Adresse erhhen
		dec R11			; Restzahl Befehlsbytes
		jr Z, ES_PPL12		;  --> fertig

ES_PPL13	call ES_PrMemByt	; nchstes Befehlsbyte drucken
		ld 5Ah, #' '
		call ZBS4		; Leerzeichen drucken
		djnz R11, ES_PPL13

		dec 5Bh			; Cursor auf letztes Zeichen
ES_PPL12	jp NLN			;  und neue Zeile anfangen



;****************************************************************************
; Magnetband-Ein/Ausgaberoutinen
;****************************************************************************


;-----------------------------------------------------------------------------
; Magnetbandausgabe im KC85-Format. RP = #%60
; Startadresse, Endadresse und Programmname werden abgefragt
; Bei ungltiger Endadresse wird der Wert aus %52/53 genommen.
;-----------------------------------------------------------------------------

	assume RP:060h
ES_Save1	call CLS
		call ES_GetHexWrd	; Startadresse holen
		jr C, ES_Save1		; ungltig --> nochmal!
		ld 50h, R15		; Startadresse in %51/50
		ld 51h, R14
		inc 5Bh			; Leerzeichen
		call ES_GetHexWrd	; Endadresse holen
		jr C, ES_Save2		; ungltig -->
		ld 52h, R14		; Endadresse in %52/53
		ld 53h, R15
ES_Save2	ld R14, 52h
		ld 52h, 53h		;  %52 und %53 tauschen
		ld 53h, R14

; Programmnamen abfragen
ES_Save		ld  5Bh, #20h		; Cursor in dritte Zeile
ES_Save3	call TBS		; auf Tastendruck warten
		cp 5Ah, #0Dh		; Enter?
		jr NZ, ES_Save3		; nein, nchstes Zeichen

; der Programmname steht jetzt im Bildschirmspeicher ab %FD20
		di     			; Video aus
		ld R0, #0FCh		; Pufferadresse
		ld 03h, R0
		ld R1, #0
		ld R2, #0FDh		; Anfang des Programmnamens
		ld R3, #020h
		ld R4, #11
ES_Save4	lde R5, @RR2		; Zeichen holen
		lde @RR0, R5		; und in den Puffer schreiben
		inc R1
		inc R3
		djnz R4, ES_Save4

; restlichen Puffer lschen
ES_Save5	lde @RR0, R4		; Null
		inc R1
		jr pl, ES_Save5		; bis %FC80 auffllen

		ld R5, #2
		ld R1, #10h
		lde @RR0, R5		; %FC10 = 02
		inc R1
		ld R5, #50h		; Start- und Endadresse LSB.MSB (!)
		ld R4, #4
ES_Save6	ldei @RR0, @r5		;  einragen ab %FC11
		djnz R4, ES_Save6

		ld R1, #0		; Zeiger wieder auf Anfang des Puffers
		ld R6, #1		; Blocknummer 1
		ld R12, #32		; Lnge Vorton
		call WriteBlock		; wegschreiben
		ld R0, 51h		; Programmstart
		ld R1, 50h
		ld R2, 53h		; Programmende
		ld R3, 52h
ES_Save8	inc R6			; nchster Block
		ld R4, R2
		ld R5, R3
		sub R5, R1		; Restlnge bestimmen
		sbc R4, R0
		jr NC, ES_Save7		; noch was da --> weiter
		ret    			; sonst fertig

; nchster Block
ES_Save7	sub R5, #128
		sbc R4, R12
		jr NC, ES_Save9		; letzter Block?
		ld R6, #255		;  ja, Blocknummer 255
ES_Save9	ld R13, #160
		call WriteBlock		; sonst wegschreiben
		jr ES_Save8


; ein Bit zum Magnetband ausgeben
; Zeit in R8 warten, P3.6 toggeln, Zeit in R8 warten, P3.6 toggeln, 

WriteBit	ld R4, #0
		ld R5, R8		; Wartezeit
WriteBit1	decw RR4		; warten
		jr NZ, WriteBit1
		xor P3, #64		; P3.6 toggeln
		ld R5, R8
WriteBit2	decw RR4		; warten
		jr NZ, WriteBit2
		xor P3, #64		; P3.6 toggeln
		ret    


; eine Flanke vom Magnetband erkennen
; R11 enthlt die Zeit seit dem letzen Zustandswechsel,
; R4 den neuen Zustand

	assume RP:040h
ReadBit		ld R11, #255		; Zeitzhler
ReadBit1	ld R12, #250		; Entprellzhler

ReadBit2	inc R11			; Zeitzhler inkrementieren
		ld R8, P3
		and R8, #TapeInMask	; P3.0 maskieren
		cp R8, R4		; gendert?
		jr Z, ReadBit1		;  nein, weiter warten

		inc R12			; entprellen
		jr NZ, ReadBit2
		ld R4, R8		; neuen Zustand merken
		ret    


; ein Byte vom Magnetband lesen
; Startbit abwarten, dann 8 Datenbits lesen und in R9 zurckliefern

ReadByte	call   ReadBit		; erste Flanke vom Startbit suchen
		call   ReadBit		; zweite Flanke suchen

ReadByt2	ld     R10, #8		; 8 Datenbit
ReadByt1	call   ReadBit		; erste Flanke suchen
		call   ReadBit		; zweite Flanke suchen
		cp     R11, #22		; langes oder kurzes bit?
		ccf    
		rrc    R9		; Zustand in R9 einbauen
		djnz   R10, ReadByt1	; bis Byte komplett
		ret    


; 128 byte Datenblock zum Magnetband ausgeben. Erst einen Vorspann aus 1er,
; (Lnge RR12), dann 1 byte Blocknummer (R6), 128 byte Nutzdaten (@RR0), 
; 1 byte Prfsumme

	assume RP:060h
WriteBlock	ld R8, #74		; Wartezeit fr 1 setzen
		call WriteBit		; Bit ausgeben
		decw rr12		; bis Vorspann fertig
		jr NZ, WriteBlock

		ld R9, R6		; Blocknummer
		call WriteByte		; auf Band schreiben
		ld R13, #128		; Blocklnge
		ld R7, #0		; Prfsumme initialisieren

WriteBlk1	lde R9, @RR0		; Byte aus Puffer holen
		incw RR0
		call WriteByte		; auf Band schreiben
		add R7, R9		; Prfsumme anpassen
		djnz R13, WriteBlk1	; bis Puffer abgearbeitet
		ld R9, R7		; Prfsumme wegschreiben

; das Byte in R9 zum Magnetband ausgeben.
; ein Startbit Lnge 148, dann 8 Datenbits Lnge 74 (1) bzw 37 (0)

WriteByte	ld R8, #148		; Wartezeit setzen
		call WriteBit		; Startbit ausgeben
		ld R10, #8		; 8 Datenbit
WriteByt2	rr R9
		ld R8, #74		; Wartezeit fr 1
		jr C, WriteByt1		; bit ist doch 1?
		ld R8, #37		;  sonst Wartezeit fr 0
WriteByt1	call WriteBit
		djnz R10, WriteByt2
		ret   
 

; Datenblock vom Magnetband lesen. Zuerst wird ein Vorspann mit einer Mindestlnge
; von 40 bit (80 Flusswechseln) gesucht, dann die Blocknummer gelesen (%5D), dann
; 128 Datenbytes (RR0) und die Prfsumme. 
; Falls die Blocknummer nicht der erwarteten entspricht (255 fr Programmende) oder
; ein Prfsummenfehler auftritt (Block 1 hat keine), dann wird der Anfang des
; gelesenen Blocks am Bildschirm ausgegeben und LOAD beendet.

	assume RP:040h
ReadBlock	ld R13, #80		; Mindestlnge Vorspann
		ld R14, #22		; 1 oder 0 ?
		ld R15, #42		; Startbit oder 1 ?

ReadBlk1	call RB_ChkBit		; Vorspann-Bit prfen
		jr C, ReadBlock		; zu kurz oder zu lang --> nochmal!
		djnz R13, ReadBlk1	; nchstes Bit

		ld R14, #42		; Startbit oder 1?
		ld R15, #106		; zu lang oder Startbit?

ReadBlk2	call RB_ChkBit		; haben wir ein Startbit?
		jr C, ReadBlk2		; nein, weitersuchen
		call RB_ChkBit		; hat die zweite Hlfte auch die richtige Lnge?
		jr C, ReadBlock		;  nein --> neu versuchen

		call ReadByt2		; Rest des ersten Datenbytes lesen
		ld 5Dh, R9		; Blocknummer in %5D
		ld R13, #128		; 128 Datenbytes
		ld R7, #0		; Prfsumme initialisieren
		inc R6			; nchster erwarteter Block
ReadBlk3	call ReadByte		; nchstes Byte einlesen
		lde @RR0, R9		; und in Puffer schreiben
		incw   RR0
		add R7, R9		; Prfsumme anpassen
		djnz R13, ReadBlk3	; bis Block komplett

		call ReadByte		; Prfsumme lesen

		ld 5Bh, #070h		; Cursor in unterste Zeile
		call ES_PrHex		; Blocknummer am Bildschirm ausgeben
		cp R7, R9		; Prfsumme ok?
		jr NZ, ReadBlk4		;  nein --> fertig oder Spezialblock
		cp R6, 5Dh		; passt Blocknummer in die Folge?
		jr Z, ReadBlk5		;  ja --> zurck zur Laderoutine

ReadBlk4	jp RB_PrintFile		; Programmname drucken

RB_ChkBit	call ReadBit		; nchste Flanke suchen
		cp R11, R14		; zu kurz?
		jr C, ReadBlk5		;  --> Ende
		cp R15, R11		; zu lang?
ReadBlk5	ret  


;-----------------------------------------------------------------------------
; Magnetbandeingabe im KC85-Format. RP = #%60  
; Zieladresse wird abgefragt. Falls ungltige Angabe wird der Wert aus 
; Block 1 genommen
;-----------------------------------------------------------------------------

ES_Load1	clr 06h			; BASIC-Start undefiniert
		dec 5Bh
		call ES_GetHexWrd	; Startadresse abfragen (in %6E/6F)
		di     			; Video aus
		jr C,ES_Load		; keine Angabe --> los
		ld  06h, R14		; sonst BASIC-Start auf Startadresse setzen
		ld  07h, R15

ES_Load		srp #040h
	assume RP:040h
		ld R0, #0FCh		; Pufferadresse
		ld R1, #0
		ld R6, #0		; erster erwarteter Block -1
		call ReadBlock		; lesen
		ld R2, #0FCh
		ld R3, #11h		; Startadresse
		ld R0, 06h		; BASIC-Start holen
		ld R1, 07h
		or R0, R0		; gltig?
		jr NZ, ES_Load2		;  ja --> Wert behalten
		lde R1, @RR2		; sonst Wert aus Puffer holen
		inc R3
		lde R0, @RR2

ES_Load2	call ReadBlock		; restliche Blcke lesen
		jr ES_Load2


	if $-ES_Begin>ROMsize
		error "\aROM size exceeded !!!"
	elseif
		db ROMsize-($-ES_Begin) dup(0FFh)	; pad with $FF's
	endif
	end
